@@ -1,5 +1,5 @@
-from src.simcalibration.utils.Data import Data
-from src.simcalibration.dg_models.DGModel import DGModel
+from simcalibration.utils.Data import Data
+from simcalibration.dg_models.DGModel import DGModel
 
 
 class Bnlearner(DGModel):
