@@ -1,6 +1,6 @@
-from ml_models.MachineLearner import MachineLearner
+from simcalibration.ml_models.MachineLearner import MachineLearner
 import pandas as pd
-from src.simcalibration.utils.Data import Data
+from simcalibration.utils.Data import Data
 
 
 class SklearnModel(MachineLearner):
