@@ -1,16 +1,21 @@
 import pandas as pd
+import os
 from typing import List, Callable
 
 from numpy import mean, int64
 from sklearn.model_selection import KFold
 
-from src.simcalibration.dg_models.DGModel import DGModel
-from ml_models.MachineLearner import MachineLearner
-from src.simcalibration.utils.Data import Data
+from simcalibration.dg_models.DGModel import DGModel
+from simcalibration.ml_models.MachineLearner import MachineLearner
+from simcalibration.utils.Data import Data
 import numpy as np
 import rpy2.robjects as robjects
 from rpy2.robjects import pandas2ri
+from scipy.spatial.distance import jensenshannon
+from sklearn.metrics import confusion_matrix
 
+figuredirname = os.path.join(os.getcwd(), "results")
+os.makedirs(figuredirname, exist_ok=True)
 pandas2ri.activate()
 
 ########################################################################################################################
@@ -68,15 +73,15 @@ class Evaluator:
                     btstrp_results[ml_model.name][score_name].append(result_per_btsrtp_per_model[score_name])
         return btstrp_results
 
-    def meta_simulate(self, dg_model_real: DGModel, n_learning: int, n_train: int,n_test: int, n_true_repetitions: int, n_practitioner_repititions: int, n_sl_repititions: int, kfolds: int):
+    def meta_simulate(self, dg_model_real: DGModel, n_learning: int, n_train: int,n_test: int, n_true_repetitions: int, n_practitioner_repetitions: int, n_sl_repetitions: int, kfolds: int):
         """
         :param dg_model_real: a real-world DagSim model
         :param n_learning: number of samples to draw from n_train to use in learners
         :param n_train: number of samples to generate for training
         :param n_test: number of samples to generate for testing
         :param n_true_repetitions: number of times to repeat benchmarking in the true real-world
-        :param n_practitioner_repititions: number of times to repeat benchmarking in a limited real-world
-        :param n_sl_repititions: number of times to repeat benchmarking in the learner
+        :param n_practitioner_repetitions: number of times to repeat benchmarking in a limited real-world
+        :param n_sl_repetitions: number of times to repeat benchmarking in the learner
         :param kfolds: number of k folds to perform in cross-validation
         :return: results: benchmarks in the shape {model_name: {learner: {score_name: list_of_score_values], ...}, ...}, ...}
         """
@@ -86,9 +91,11 @@ class Evaluator:
         list_of_npractitioner_cv_accuracies = {method.name: [] for method in self.ml_models}
         list_of_nsl_accuracies = {learner.SLClass: {method.name: [] for method in self.ml_models} for learner in self.dg_models}
         list_of_nsl_cv_accuracies = {learner.SLClass: {method.name: [] for method in self.ml_models} for learner in self.dg_models}
+        list_of_dag_fidelity = {learner.SLClass: [] for learner in self.dg_models}
+        list_of_distribution_fidelity = {learner.SLClass: [] for learner in self.dg_models}
         # Obtain true and practitioner accuracies
         dg_metrics, _, _, list_of_ntrue_cv_accuracies = self._get_performance_by_repetition_return_data(dg_model_real, n_train + n_test,0.5, n_true_repetitions, kfolds=kfolds)
-        limited_dg_metrics, train_data_list, test_data_list, list_of_npractitioner_cv_accuracies = self._get_performance_by_repetition_return_data(dg_model_real, n_train + n_test,0.5, n_practitioner_repititions, kfolds=kfolds)
+        limited_dg_metrics, train_data_list, test_data_list, list_of_npractitioner_cv_accuracies = self._get_performance_by_repetition_return_data(dg_model_real, n_train + n_test,0.5, n_practitioner_repetitions, kfolds=kfolds)
 
         # Store accuracies for true and practitioner models
         for ml in dg_metrics:
@@ -96,9 +103,14 @@ class Evaluator:
             list_of_npractitioner_accuracies[ml] = limited_dg_metrics[ml]['balanced_accuracy_score']
 
         # Begin practitioner repetitions
-        for practitioner_rep in range(0, n_practitioner_repititions):
+        for practitioner_rep in range(0, n_practitioner_repetitions):
             print("Practitioner repetition: ", practitioner_rep)
+            # get canonical column order from the real training DataFrame that will be presented to bnlearn
+            canonical_order = list(train_data_list[practitioner_rep].all.columns)
 
+            # force the ground-truth DagsimModel to use canonical order for true adjacency
+            if hasattr(dg_model_real, "set_var_order"):
+                dg_model_real.set_var_order(canonical_order)
 
             for dg_model in self.dg_models:
                 print("Learner: ", dg_model.SLClass)
@@ -109,8 +121,28 @@ class Evaluator:
                 sl_temp_scores = {ml.name: [] for ml in self.ml_models}
 
                 # Perform SL repetitions and capture metrics
-                for sl_rep in range(0, n_sl_repititions):
-                    sl_dg_metrics, sl_train_data, sl_test_data, current_cv_metrics = self._evaluate_bnlearn_dg_model(dg_model=dg_model,learning_data_real=train_data_list[practitioner_rep],n_learning=n_learning,n_train=n_train,n_test=n_test,SLClass=dg_model.SLClass, kfolds=kfolds)
+                for sl_rep in range(0, n_sl_repetitions):
+                    print("SL repetition: ", sl_rep)
+                    sl_dg_metrics, sl_learning_data, sl_test_data, current_cv_metrics = self._evaluate_bnlearn_dg_model(dg_model=dg_model,learning_data_real=train_data_list[practitioner_rep],n_learning=n_learning,n_train=n_train,n_test=n_test,SLClass=dg_model.SLClass, kfolds=kfolds)
+                    # Compute DAG structural fidelity
+                    true_dag = dg_model_real.dag  # adjacency matrix or graph representation
+                    learned_dag = dg_model.learned_dag  # adjacency matrix or graph representation from bnlearn
+                    dag_fid = self.evaluate_structure(
+                        true_adj=dg_model_real.dag,
+                        learned_adj=dg_model.learned_dag,
+                        true_vars=getattr(dg_model_real, "var_names", None),
+                        learned_vars=getattr(dg_model, "var_names", None)
+                    )
+                    list_of_dag_fidelity[dg_model.SLClass].append(dag_fid)
+
+                    # Compute distribution fidelity
+                    real_data = train_data_list[practitioner_rep]
+                    synthetic_data = sl_test_data  # or sl_test_data
+                    real_df = pd.concat([real_data.X, real_data.y], axis=1)
+                    synth_df = pd.concat([synthetic_data.X, synthetic_data.y], axis=1)
+                    dist_fid = self.evaluate_distribution(real_df, synth_df)
+                    list_of_distribution_fidelity[dg_model.SLClass].append(dist_fid)
+
                     for ml in self.ml_models:
                         sl_temp_scores[ml.name].append(current_cv_metrics[dg_model.SLClass][ml.name]['balanced_accuracy_score'])
                     for score_name in repetition_results.index:
@@ -141,7 +173,23 @@ class Evaluator:
         pd.DataFrame(list_of_npractitioner_cv_accuracies).to_csv("npractitioner_cv_accuracies.csv")
         pd.DataFrame(list_of_nsl_accuracies).to_csv("nsl_accuracies.csv")
         pd.DataFrame(list_of_nsl_cv_accuracies).to_csv("nsl_cv_accuracies.csv")
-        return [list_of_ntrue_accuracies, list_of_npractitioner_accuracies, list_of_nsl_accuracies, n_true_repetitions, n_practitioner_repititions, n_sl_repititions, self.dg_models, self.ml_models, list_of_ntrue_cv_accuracies, list_of_npractitioner_cv_accuracies, list_of_nsl_cv_accuracies]
+        dag_rows = []
+        for sl, records in list_of_dag_fidelity.items():
+            for rep_idx, rec in enumerate(records):
+                row = {"SL": sl, "rep": rep_idx}
+                row.update(rec)
+                dag_rows.append(row)
+        pd.DataFrame(dag_rows).to_csv("dag_fidelity.csv", index=False)
+        dist_rows = []
+        for sl, records in list_of_distribution_fidelity.items():
+            for rep_idx, rec in enumerate(records):
+                row = {"SL": sl, "rep": rep_idx, "JS_divergence": rec["JS_divergence"]}
+                # flatten per-column divergences
+                for col, js_val in rec["per_column"].items():
+                    row[f"JS_{col}"] = js_val
+                dist_rows.append(row)
+        pd.DataFrame(dist_rows).to_csv("distribution_fidelity.csv", index=False)
+        return [list_of_ntrue_accuracies, list_of_npractitioner_accuracies, list_of_nsl_accuracies, n_true_repetitions, n_practitioner_repetitions, n_sl_repetitions, self.dg_models, self.ml_models, list_of_ntrue_cv_accuracies, list_of_npractitioner_cv_accuracies, list_of_nsl_cv_accuracies, list_of_dag_fidelity,list_of_distribution_fidelity]
 
     def _evaluate_ml_model(self, ml_model: MachineLearner, test_data: Data):
         """
@@ -201,6 +249,12 @@ class Evaluator:
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
                         list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)          
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "tabu":
@@ -222,7 +276,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "rsmax2":
@@ -244,7 +303,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "mmhc":
@@ -266,7 +330,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "h2pc":
@@ -288,7 +357,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "gs":
@@ -310,7 +384,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "iamb":
@@ -332,7 +411,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "fast.iamb":
@@ -354,7 +438,12 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
         elif SLClass == "iamb.fdr":
@@ -376,22 +465,56 @@ class Evaluator:
                         fit = bn.fit(my_bn, learning_data_real)
                         training_output = rbn(fit, n_train, keep.fitted = TRUE)
                         testing_output = rbn(fit, n_test, keep.fitted = TRUE)
-                        list_output <- list(na.omit(training_output), na.omit(testing_output))
+                        A <- amat(my_bn)            
+                        nodes <- colnames(learning_data_real)
+                        return(list(train = na.omit(training_output),
+                                    test  = na.omit(testing_output),
+                                    adj   = A,
+                                    nodes = nodes))
                         }
                         ''')
 
         bn_hc = robjects.r['bn_learn']
         bn_train_output = bn_hc(learning_data_real, n_train, n_test)
 
-        train_data = np.array(bn_train_output[0], dtype=int64)
-        test_data = np.array(bn_train_output[1], dtype=int64)
-        X = pd.DataFrame(train_data[:, :-1])
-        y = pd.Series(train_data[:, -1], name="Y")
+        # --- Extract train/test arrays (as before) ---
+        try:
+            train_array = np.array(bn_train_output.rx2('train'), dtype=np.int64)
+            test_array = np.array(bn_train_output.rx2('test'), dtype=np.int64)
+        except Exception:
+            # fallback if bn_learn() wasn’t updated to return named elements
+            train_array = np.array(bn_train_output[0], dtype=np.int64)
+            test_array = np.array(bn_train_output[1], dtype=np.int64)
+
+        # --- Build Data objects (same as before) ---
+        X = pd.DataFrame(train_array[:, :-1])
+        y = pd.Series(train_array[:, -1], name="Y")
         train_data = Data(name="train", X=X, y=y)
-        X = pd.DataFrame(test_data[:, :-1])
-        y = pd.Series(test_data[:, -1], name="Y")
+
+        X = pd.DataFrame(test_array[:, :-1])
+        y = pd.Series(test_array[:, -1], name="Y")
         test_data = Data(name="test", X=X, y=y)
 
+        # --- NEW: extract adjacency + nodes (if returned by bn_learn) ---
+        try:
+            A_r = bn_train_output.rx2('adj')  # R matrix from amat()
+            bn_nodes = [str(n) for n in bn_train_output.rx2('nodes')]
+            A = np.array(A_r, dtype=np.int64)
+
+            canonical_order = list(learning_data_real.columns)  # pandas col order
+            index = [bn_nodes.index(n) for n in canonical_order]
+            A_reordered = A[np.ix_(index, index)]
+
+            # attach to learner for later fidelity evaluation
+            dg_model.learned_dag = A_reordered
+            dg_model.var_names = canonical_order
+
+        except Exception as e:
+            print(f"[warn] adjacency/nodes not available from bn_learn: {e}")
+            dg_model.learned_dag = None
+            dg_model.var_names = None
+
+        # --- CV and ML evaluation (unchanged) ---
         kf = KFold(n_splits=kfolds, shuffle=True, random_state=42)
         cv_metrics = {}
 
@@ -409,6 +532,7 @@ class Evaluator:
                 ml_model.learn(fold_train_data)
                 fold_scores = self._evaluate_ml_model(ml_model, fold_val_data)
                 model_fold_metrics.append(fold_scores)
+
             avg_model_metrics = {
                 metric: np.mean([fold[metric] for fold in model_fold_metrics])
                 for metric in model_fold_metrics[0]
@@ -500,3 +624,91 @@ class Evaluator:
         train_data = dg_model_real.generate(int(n_samples * tr_frac), self.outcome_name)
         test_data = dg_model_real.generate(int(n_samples * (1 - tr_frac)), self.outcome_name)
         return train_data, test_data
+
+    @staticmethod
+    def evaluate_structure(true_adj, learned_adj, true_vars=None, learned_vars=None):
+        """
+        Align learned_adj to true_adj by variable names (if provided) and compute SHD + precision/recall/F1.
+        true_adj, learned_adj: numpy arrays (0/1)
+        true_vars, learned_vars: list of variable names corresponding to rows/cols (optional)
+        """
+        # If either adj is None -> return NAs
+        if true_adj is None or learned_adj is None:
+            return {"SHD": np.nan, "Precision": np.nan, "Recall": np.nan, "F1": np.nan}
+
+        # If names provided, reorder learned_adj to match true_vars
+        if true_vars is not None and learned_vars is not None:
+            # compute index mapping for intersection
+            intersect = [v for v in true_vars if v in learned_vars]
+            if len(intersect) == 0:
+                return {"SHD": np.nan, "Precision": np.nan, "Recall": np.nan, "F1": np.nan}
+            t_idx = [true_vars.index(v) for v in intersect]
+            l_idx = [learned_vars.index(v) for v in intersect]
+            true_sub = true_adj[np.ix_(t_idx, t_idx)]
+            learned_sub = learned_adj[np.ix_(l_idx, l_idx)]
+        else:
+            # use provided arrays as-is (must be same shape)
+            true_sub = true_adj
+            learned_sub = learned_adj
+
+        # Ensure same shape
+        if true_sub.shape != learned_sub.shape:
+            min_n = min(true_sub.shape[0], learned_sub.shape[0])
+            true_sub = true_sub[:min_n, :min_n]
+            learned_sub = learned_sub[:min_n, :min_n]
+
+        # binary matrices (0/1)
+        A = (true_sub != 0).astype(int)
+        B = (learned_sub != 0).astype(int)
+
+        # symmetric difference -> SHD
+        shd = int(np.sum(np.abs(A - B)))
+
+        # compute tp/fp/fn (directed edges)
+        tp = int(np.sum((A == 1) & (B == 1)))
+        fp = int(np.sum((A == 0) & (B == 1)))
+        fn = int(np.sum((A == 1) & (B == 0)))
+
+        precision = tp / (tp + fp + 1e-12) if (tp + fp) > 0 else 0.0
+        recall = tp / (tp + fn + 1e-12) if (tp + fn) > 0 else 0.0
+        f1 = 2 * precision * recall / (precision + recall + 1e-12) if (precision + recall) > 0 else 0.0
+
+        return {"SHD": shd, "Precision": precision, "Recall": recall, "F1": f1}
+
+    @staticmethod
+    def evaluate_distribution(real_data: pd.DataFrame, synthetic_data: pd.DataFrame, n_bins=20):
+        """
+        Compute per-column JS divergence between real_data and synthetic_data.
+        - For categorical (dtype == object or int with few unique values) use value counts.
+        - For continuous-ish use histogram with n_bins.
+        Returns mean JS divergence and column-wise values.
+        """
+        cols = [c for c in real_data.columns if c in synthetic_data.columns]
+        col_js = {}
+        for col in cols:
+            real_col = real_data[col].dropna()
+            synth_col = synthetic_data[col].dropna()
+
+            # detect categorical (small number of unique values or object dtype)
+            if real_col.dtype == object or (real_col.nunique() <= 10):
+                # use value counts over union of categories
+                cats = sorted(list(set(real_col.unique()) | set(synth_col.unique())))
+                p = np.array([(real_col == c).sum() for c in cats], dtype=float)
+                q = np.array([(synth_col == c).sum() for c in cats], dtype=float)
+            else:
+                # continuous -> histogram
+                combined = np.concatenate([real_col.values, synth_col.values])
+                bins = np.histogram_bin_edges(combined, bins=n_bins)
+                p, _ = np.histogram(real_col.values, bins=bins)
+                q, _ = np.histogram(synth_col.values, bins=bins)
+
+            # normalize
+            p = p / (p.sum() + 1e-12)
+            q = q / (q.sum() + 1e-12)
+
+            js = float(jensenshannon(p, q))  # scalar
+            col_js[col] = js
+
+        mean_js = float(np.mean(list(col_js.values()))) if col_js else np.nan
+        return {"JS_divergence": mean_js, "per_column": col_js}
+
